#include <algorithm>
#include <atomic>
#include <cassert>
#include <chrono>
#include <coroutine>
#include <filesystem>
#include <fstream>
#include <functional>
#include <iostream>
#include <list>
#include <ranges>
#include <unordered_set>
namespace ipfix {
       struct error {};
       struct logic_error : std::logic_error, error {};
       struct runtime_error : std::runtime_error, error {};
       struct out_of_range : std::out_of_range, error {};
       struct length_error : std::length_error, error {};
       namespace fmt = ::std;
       }
        namespace ipfix {
       template <typename CharT> struct base_formatter {
        constexpr auto parse(fmt::basic_format_parse_context<CharT> const & ctx) {
         if (ctx.begin() == ctx.end() || *ctx.begin() == '}') 
          return ctx.begin();
        
         throw fmt::format_error("invalid format");
       }
      };
       class opt_formatter {
       public:   constexpr auto parse(fmt::basic_format_parse_context<char> const & ctx) {
         auto i = ctx.begin();
         for (;
     i != ctx.end();
     ++i) ;
         return i;
       }
      };
       
       }
        namespace ipfix {
       namespace date = ::std::chrono;
       }
        namespace ipfix::chrono {
       namespace sc = std::chrono;
       namespace detail {
      struct ipfix_clock {
       using base_clock = date::system_clock;
     };
      }
       template <typename T> concept is_ipfix_clock =   date::is_clock_v<T> && std::derived_from<T, detail::ipfix_clock>;
       struct utc_clock : detail::ipfix_clock {
        using rep = base_clock::rep;
        using period = base_clock::period;
        using duration = base_clock::duration;
        using time_point = sc::time_point<utc_clock, duration>;
        template <typename Duration>   using time = sc::time_point<utc_clock, Duration>;
        static constexpr bool is_steady = base_clock::is_steady;
        static time_point now() noexcept ;
      };
       template <typename Duration> using utc_time = utc_clock::time<Duration>;
       }
        template <> struct ipfix::date::clock_time_conversion <ipfix::chrono::utc_clock::base_clock,  ipfix::chrono::utc_clock> {
         template <typename Duration>   constexpr auto   operator()(ipfix::chrono::utc_time<Duration> const & tp)     const noexcept   {
          using base_clock = ipfix::chrono::utc_clock::base_clock;
          return std::chrono::time_point<base_clock, Duration>{
           tp.time_since_epoch()};
        }
       };
        namespace ipfix::chrono {
       namespace detail {}
       class ntp_clock : public detail::ipfix_clock {
        using ntp_duration =     sc::duration<std::uint64_t, std::ratio<1, 0x100000000L>>;
       public:   using rep = ntp_duration::rep;
        using duration = ntp_duration;
        using time_point = sc::time_point<ntp_clock>;
      };
       using dateTimeSeconds_t = utc_time<sc::duration<std::uint32_t>>;
       using dateTimeMilliseconds_t =   utc_time<sc::duration<std::uint64_t, std::milli>>;
       struct dateTimeMicroseconds_t : ntp_clock::time_point {
        using base = ntp_clock::time_point;
        constexpr dateTimeMicroseconds_t(duration const & d) :     base{
    duration{
   d.count() | ~static_cast<rep>(0x7ff)}
    }
        {}
      };
       using dateTimeNanoseconds_t = ntp_clock::time_point;
       template <typename T> constexpr bool is_time_point = false;
       template <typename C, typename D> constexpr bool is_time_point<sc::time_point<C, D>> = true;
       template <> constexpr bool is_time_point<dateTimeMicroseconds_t> = true;
       template <typename T> concept TimePoint = is_time_point<T>;
       template <typename T> concept IpfixTimePoint = TimePoint<T> && is_ipfix_clock<typename T::clock>;
       }
        namespace ipfix::chrono::detail {
       template <IpfixTimePoint TP> using base_clock_cast_rv_t =   decltype(clock_cast<typename TP::clock::base_clock>(std::declval<TP>()));
       template <TimePoint TP, typename Duration> struct formatter;
       template <typename Char, IpfixTimePoint TP> struct formatter<TP, Char> : fmt::formatter<base_clock_cast_rv_t<TP>> {
        using base = fmt::formatter<base_clock_cast_rv_t<TP>>;
        auto format(TP const & tp, auto & ctx) const {
         auto ctp = clock_cast<typename TP::clock::base_clock>(tp);
         return base::format(ctp, ctx);
       }
      };
       }
        template <typename Char, ipfix::chrono::is_ipfix_clock Clock,           typename Duration> struct ipfix::fmt::formatter<   std::chrono::time_point<Clock, Duration>, Char> : ipfix::chrono::detail::formatter<   std::chrono::time_point<Clock, Duration>, Char> {};
        namespace ipfix::types {
       using octet_t = std::uint8_t;
       using length_t = std::uint16_t;
       }
        namespace ipfix {}
        namespace ipfix::types {
       struct IsNamedTypeParam {};
       template <typename T, typename Param,           template <typename> class... Abilities> class NamedType :   public IsNamedTypeParam,   public Abilities<NamedType<T, Param, Abilities...>>... {
        struct testBase     : public Abilities<NamedType>...   {};
        static constexpr bool has_value_ =     requires (testBase x) {
     x.value_0;
     };
        static constexpr bool has_default_value_ =     requires {
     testBase::default_value_0;
     };
        T val_;
       public:   using value_type = T;
        explicit constexpr NamedType(T const & v) : val_{
    v}
      {}
        constexpr NamedType() requires (!has_default_value_)     : val_{}
        {}
         constexpr T const & raw_value() const noexcept   {
     return val_;
     }
         constexpr T const & value() const noexcept   requires (!has_value_)   {
     return raw_value();
     }
      };
       template <typename T, template <typename> class Base> struct CRTP {
         constexpr T const & cast() const noexcept {
         return static_cast<T const &>(*this);
       }
      };
       template <typename T> struct Addable : CRTP<T, Addable> {};
       template <typename T> struct EqComparable : CRTP<T, EqComparable> {};
       template <typename T> struct RangeComparable : CRTP<T, RangeComparable> {};
       template <typename T> struct Comparable : EqComparable<T>, RangeComparable<T> {};
       template <typename Value> struct AutoConvertible {
        template <typename T>   struct impl : CRTP<T, impl>   {
          constexpr operator Value() const {
          return this->cast().raw_value();
        }
       };
      };
       template <typename Value> struct ChunkReadable {
        template <typename T>   struct impl : CRTP<T, impl>   {};
      };
       template <typename Value, Value Default> struct Defaulted ;
       template <typename T> struct Hashable : CRTP<T, Hashable> {
         auto hash() const noexcept {
         return std::hash<std::decay_t<decltype(this->cast().raw_value())>>
    ();
       }
      };
       
       
      template <typename T> static constexpr bool has_extras = requires {
     T::extras;
     };
      template <typename T> concept extras_container =   ((!has_extras<T>     || 0));
      template <typename T> concept IntegralOrEnum = std::integral<T> || std::is_enum_v<T>;
      template <typename T> struct UV;
      template <std::integral T> struct UVT ;
      template <typename T> using UV_t = typename UV<T>::type;
      template <typename T> concept names_container = IntegralOrEnum<typename T::type>   && requires {
       {
    T::names }
     -> std::ranges::sized_range;
     }
      && extras_container<T>;
      template <names_container Names> struct name_container_helper ;
      
      
       template <names_container Names> struct Titled {
        template <typename T>   class impl : public CRTP<T, impl>   {};
      };
       template <typename T> concept IsNamedType = std::derived_from<T, IsNamedTypeParam>;
       template <typename T> concept IsHashable =   requires (T const & t) {
      {
     t.hash() }
      -> std::integral;
      };
       }
        namespace std {
       template <typename T> requires (ipfix::types::IsHashable<T>           && requires (T const & t) {
      {
     t.hash() }
      -> std::integral;
      }
      ) struct hashT ;
       }
        template <typename T, typename Char> requires (ipfix::types::IsNamedType<T>           && std::integral<typename T::value_type>) struct ipfix::fmt::formatter<T, Char>   : ipfix::fmt::formatter<typename T::value_type, Char> {};
        template <typename T, typename Char> requires (ipfix::types::IsNamedType<T>           && std::integral<typename T::value_type>           && requires (T t) {
       {
      t.to_name0 }
       -> std::same_as<std::string>;
       }
       ) struct ipfix::fmt::formatter<T, Char>   : ipfix::fmt::formatter<std::string, Char> {};
        namespace ipfix::types {
       template<typename T> concept OctetLike =   sizeof(T) == sizeof(octet_t)   && sizeof(T *) == sizeof(octet_t *)   && alignof(T) == alignof(octet_t)   && std::has_unique_object_representations_v<T>;
       template<typename T> concept OctetContiguousRange =   std::ranges::contiguous_range<T>   && std::ranges::sized_range<T>   && OctetLike<std::ranges::range_value_t<T>>;
       using Length = NamedType<length_t, struct LengthParam,                          Addable, Comparable,                          AutoConvertible<length_t>::impl,                          ChunkReadable<length_t>::impl>;
       template <typename T> struct ElementAdditions : CRTP<T, ElementAdditions> {};
       using ElementID =   NamedType<std::uint16_t, struct InfoElementIDParam,             Comparable, ChunkReadable<std::uint16_t>::impl,             ElementAdditions>;
       using EnterpriseNumber =   NamedType<std::uint32_t, struct EnterpriseNumberParam,             ChunkReadable<std::uint32_t>::impl, Comparable>;
       class InfoElementID {
        static constexpr EnterpriseNumber NO_ENTERPRISE_NUMBER{};
        EnterpriseNumber enterprise_number_;
        ElementID ie_;
       public:   
        explicit constexpr InfoElementID(     ElementID ie,     EnterpriseNumber enterprise_number = NO_ENTERPRISE_NUMBER) noexcept     : enterprise_number_{
    enterprise_number}
     , ie_{
    ie}
        {}
        bool operator==(const InfoElementID &) const noexcept = default;
        auto hash() const noexcept {
         std::size_t seed = 0;
         return seed;
       }
        
      };
       using TemplateID = NamedType<std::uint16_t, struct TemplateIDParam,                              Comparable, Hashable,                              ChunkReadable<std::uint16_t>::impl>;
       struct DataTypeNames {
        using type = std::uint8_t;
        static constexpr std::array names = {
         "octetArray",     "unsigned8",     "unsigned16",     "unsigned32",     "unsigned64",     "signed8",     "signed16",     "signed32",     "signed64",     "float32",     "float64",     "boolean",     "macAddress",     "string",     "dateTimeSeconds",     "dateTimeMilliseconds",     "dateTimeMicroseconds",     "dateTimeNanoseconds",     "ipv4Address",     "ipv6Address",     "basicList",     "subTemplateList",     "subTemplateMultiList"};
      };
       using DataType = NamedType<std::uint8_t, struct DataTypeParam, EqComparable,                            ChunkReadable<std::uint8_t>::impl,                            Titled<DataTypeNames>::impl>;
       struct DataTypeSemanticsNames {
        using type = std::uint8_t;
        static constexpr char const * names[] = {
         "default",     "quantity",     "totalCounter",     "deltaCounter",     "identifier",     "flags",     "list",     "snmpCounter",     "snmpGauge"};
      };
       using DataTypeSemantics = NamedType<   std::uint8_t, struct DataTypeSemanticsParam, EqComparable,   ChunkReadable<std::uint8_t>::impl,   Titled<DataTypeSemanticsNames>::impl>;
       struct UnitsNames {
        using type = std::uint16_t;
        static constexpr char const * names[] = {
         "none",     "bits",     "octets",     "packets",     "flows",     "seconds",     "milliseconds",     "microseconds",     "nanoseconds",     "4-octet words",     "messages",     "hops",     "entries",     "frames",     "ports",     "inferred"   };
      };
       using Units = NamedType<   std::uint16_t, struct UnitsParam, EqComparable,   Titled<UnitsNames>::impl>;
       struct StructuredDataTypeSemanticsNames ;
       }
        template <typename CharT> struct ipfix::fmt::formatter<   ipfix::types::InfoElementID, CharT>   : public ipfix::base_formatter<CharT> {};
        namespace ipfix::byteswap {}
        namespace ipfix::floatpoint {
       
      template <typename... T> struct Types ;
      template <template <typename> class Match, typename T> struct first_of {
     using type = void;
     };
      template <template <typename> class Match, typename T, typename... Rest> struct first_of<Match, Types<T, Rest...>> {
       using type =     std::conditional_t<Match<T>::value, T,                        typename first_of<Match, Types<Rest...>>::type>;
     };
      template <std::size_t size> struct matching_iec559 {
       template <typename T>   struct match {
        static constexpr bool value =       sizeof(T) == size && std::is_scalar<T>::value &&       std::numeric_limits<T>::is_iec559;
      };
     };
      template <typename T> using iec559_double32 = typename matching_iec559<4>::match<T>;
      template <typename T> using iec559_double64 = typename matching_iec559<8>::match<T>;
      using types = Types<float, double, long double>;
      using float32_t = typename first_of<iec559_double32, types>::type;
      using float64_t = typename first_of<iec559_double64, types>::type;
      template <typename T> consteval auto fp_endianness_helper() ;
      
      
       
       
       }
        namespace ipfix::convert {
       template <typename T> concept Chunk = types::OctetContiguousRange<T>;
       using std::ranges::data;
       using std::ranges::size;
       using std::ranges::empty;
       using floatpoint::float32_t;
       using floatpoint::float64_t;
       template <typename T> struct from_chunk;
       template <> struct from_chunk<std::string_view> {};
       template <> struct from_chunk<std::optional<bool>> {};
       template <types::OctetLike T> struct from_chunk<std::span<T const>> {};
       template <std::size_t N, types::OctetLike T> requires (N != std::dynamic_extent) struct from_chunk<std::span<T const, N>> {};
       template <std::integral T, bool reducable>  T from_chunk_integral_helper() ;
       template <std::integral T> struct from_chunk<T> {};
       template <chrono::TimePoint T> requires std::integral<typename T::rep> struct from_chunk<T> {};
       template <> struct from_chunk<float32_t> {};
       template <> struct from_chunk<float64_t> {
           float64_t operator()(Chunk auto const & chunk) const {
         if (chunk.size() == sizeof(float32_t)) 
          return from_chunk{}
   (chunk);
        
         std::array<types::octet_t, sizeof(float64_t)> buf;
         return std::bit_cast<float64_t>(buf);
       }
      };
       }
        namespace ipfix {
       using namespace types;
       class Chunk : public std::ranges::view_interface<Chunk> {
        std::shared_ptr<octet_t> data_;
        length_t size_ = 0;
       public:   using iterator = octet_t const *;
        Chunk() = default;
        Chunk(Chunk const & other, length_t offset) :     size_{
    static_cast<length_t>(other.size_ - offset)}
        {}
        Chunk(Chunk const & other, length_t offset, length_t size) :     size_{}
        {
         other.check_size(offset, size);
       }
         iterator begin() const noexcept ;
         iterator end() const noexcept ;
        void  shrink() ;
         void  subchunk() const & ;
         void  subchunk(length_t offset) const & ;
        template <std::integral T>    T read() const ;
        template <chrono::TimePoint T>   requires std::integral<typename T::rep>    T read() const ;
        template <typename T>   requires requires {
         typename T::chunk_read_type;
       }
         T read() const ;
        template <typename T>   requires requires  {
         {
    T::chunk_read0 }
     -> std::same_as<T>;
       }
         T read() const ;
       private:   struct empty_deleter ;
        
        Chunk & set_bounds() ;
        void check_size(length_t offset, length_t size) const ;
        
      };
       }
        namespace ipfix::utils {
       struct string_hash ;;
       template <typename... Types> using tuple_optional =   std::tuple<>;
       template <typename Ident> class InternedStrings ;
       template <typename... Args> class CallbackList {
        using fn_type = void(Args...);
        std::list<std::function<fn_type>> list;
       public:   using handle = std::ranges::iterator_t<decltype(list)>;
        template <std::invocable<Args...> F>   handle add(F && fn) {
         return list.insert(list.end(), std::function(std::move(fn)));
       }
        
      };
       template <std::size_t Size, template <std::size_t> typename Fn,           typename R, typename... Args> requires (std::regular_invocable<Fn<0>, Args...>           && std::same_as<R, std::invoke_result_t<Fn<0>, Args...>>) class IndexedCaller {
        template <std::size_t I>   static constexpr R call_fn(Args &&... args) {
         return Fn<I>{}
    (std::forward<Args>(args)...);
       }
        template <std::size_t... I>   static consteval auto fn_array(std::index_sequence<I...>) noexcept {
         return std::array{
   &call_fn<I>...};
       }
       public:   constexpr R operator()(std::size_t i, Args &&... args) {
         return fn_array(std::make_index_sequence<Size>{}
    ).at(i)       (std::forward<Args>(args)...);
       }
      };
       }
        namespace ipfix {
       using octetArray_t = std::span<octet_t const>;
       using unsigned8_t = std::uint8_t;
       using unsigned16_t = std::uint16_t;
       using unsigned32_t = std::uint32_t;
       using unsigned64_t = std::uint64_t;
       using signed8_t = std::int8_t;
       using signed16_t = std::int16_t;
       using signed32_t = std::int32_t;
       using signed64_t = std::int64_t;
       using float32_t = floatpoint::float32_t;
       using float64_t = floatpoint::float64_t;
       using boolean_t = std::optional<bool>;
       using macAddress_t = std::span<octet_t const, 6>;
       using string_t = std::string_view;
       using dateTimeSeconds_t = chrono::dateTimeSeconds_t;
       using dateTimeMilliseconds_t = chrono::dateTimeMilliseconds_t;
       using dateTimeMicroseconds_t = chrono::dateTimeMicroseconds_t;
       using dateTimeNanoseconds_t = chrono::dateTimeNanoseconds_t;
       using ipv4Address_t = std::span<octet_t const, 4>;
       using ipv6Address_t = std::span<octet_t const, 16>;
       using basicList_t = class BasicList;
       using subTemplateList_t = class SubTemplateList;
       using subTemplateMultiList_t = class SubTemplateMultiList;
       using version_t = std::uint16_t;
       using export_t = dateTimeSeconds_t;
       using sequence_t = std::uint32_t;
       using obdomain_t = std::uint32_t;
       using setid_t = std::uint16_t;
       using scope_t = std::uint16_t;
       static constexpr Length VARLEN;
       template <typename... T> struct Types {
        static constexpr std::size_t size = sizeof...(T);
      };
       class defaults {
        template <std::size_t I, typename T>   struct element;
        template <std::size_t I, typename T, typename... Rest>   struct element<I, Types<T, Rest...>> : element<I-1, Types<Rest...>> {};
        template <typename T, typename... Rest>   struct element<0, Types<T, Rest...>> {
         using type = T;
       };
       public:   using types =     Types<     octetArray_t,     unsigned8_t,     unsigned16_t,     unsigned32_t,     unsigned64_t,     signed8_t,     signed16_t,     signed32_t,     signed64_t,     float32_t,     float64_t,     boolean_t,     macAddress_t,     string_t,     dateTimeSeconds_t,     dateTimeMilliseconds_t,     dateTimeMicroseconds_t,     dateTimeNanoseconds_t,     ipv4Address_t,     ipv6Address_t,     basicList_t,     subTemplateList_t,     subTemplateMultiList_t     >;
        template <std::size_t I>   using element_t = typename element<I, types>::type;
      };
       }
        namespace ipfix {
       class Packet : public Chunk {};
       }
        namespace ipfix::generator {
       template <typename T> class Generator {
       public:   struct promise_type;
       private:   using handle = std::coroutine_handle<>;
        handle h_;
        T & get_value ;
       public:   Generator() : h_{}
      {}
        void  operator=(Generator && other) noexcept ;
        class iterator {
         Generator * rc_;
        public:     
         iterator(Generator * rc) : rc_{
   rc}
     {}
         iterator() = default;
         iterator & operator=(iterator &&) = default;
         using difference_type = std::ptrdiff_t;
         using value_type = std::remove_reference_t<T>;
         T & operator *() const ;
         void  operator++() ;
         void operator++(int) ;
         bool done() const ;
         friend bool operator==(iterator const & i, std::default_sentinel_t)       noexcept     {
          return i.done();
        }
       };
        auto begin() noexcept {
         return iterator{
   this};
       }
        auto end() const noexcept {
     return std::default_sentinel;
     }
        struct promise_type ;
      };
       }
        namespace ipfix {
       class InfoElement;
       }
        namespace ipfix::xml {
       using Ptr = std::shared_ptr<const InfoElement>;
       generator::Generator<Ptr> read_xml_infoelements(   std::string const & xml);
       }
        namespace ipfix {
       class InfoElement {
        InfoElementID id_;
        std::string name_;
        std::string_view group_;
        DataType datatype_;
        DataTypeSemantics datatype_semantics_;
        Units units_;
        std::optional<bool> reversible_;
       public:   using Ptr = std::shared_ptr<const InfoElement>;
        
        std::string const & name() const ;
        DataType datatype() const ;
        template <typename T>     requires std::constructible_from<std::string_view, T>    friend bool operator==(InfoElement const & element,                                        T const & name) noexcept   {
         return element.name_ == std::string_view{};
       }
        class Builder ;
        
        struct NameHasher ;
        
      };
       template <typename T> requires std::constructible_from<std::string_view, T>  inline bool operator==(InfoElement::Ptr const & element,                                      T const & name) noexcept ;
       template <typename T> concept InfoElementReverser =   std::invocable<T, InfoElement const &>   && requires () {
          bool{};
      };
       InfoElement::Ptr reverse_standard_ie(InfoElement const & ie) noexcept;
       class reverse_private_ie ;
       class InfoModel {
        struct InfoModelBase;
        using Ptr = std::shared_ptr<InfoModelBase>;
        Ptr data_;
       public:   InfoModel() : data_{}
      {}
        InfoModel submodel() const ;
        
        template <typename Source, InfoElementReverser Reverse>   void read_xml(Source xml, Reverse fn) {
         read_xml(xml::read_xml_infoelements(xml), fn);
       }
         InfoElement::Ptr find(InfoElementID const & id)     const noexcept   ;
       private:   inline static constinit std::atomic update_counter_ = 0;
        static InfoElement::Ptr   reverse_identity_ ;
        template <InfoElementReverser Reverse>   void read_xml(generator::Generator<InfoElement::Ptr> && gen,                 Reverse fn = reverse_identity_)   {
         for (auto & ie : gen) 
          auto reverse = fn(*ie);
        
       }
        struct InfoModelBase ;
      };
       }
        namespace ipfix::detail {
       template <typename Parent, typename Value> class iterator ;
       }
        namespace ipfix {
       class PacketView : public std::ranges::view_interface<PacketView> {};
       class TemplateManager;
       class Source {
        std::unique_ptr<TemplateManager> templates_;
       public:   using Ptr = std::shared_ptr<Source>;
        Source(InfoModel const & model)     : templates_{
    std::make_unique<TemplateManager>(model.submodel())}
        {}
        void  infomodel() const noexcept;
        void  template_map() const noexcept ;
        virtual std::string_view name() const noexcept = 0;
        virtual generator::Generator<Packet const> packets() = 0;
      };
       class FileSource : public Source {
        std::string filename_;
        PacketView packets_;
       public:   FileSource(std::filesystem::path const & fn, InfoModel const & model)     : Source(model),       filename_{
    fn.native()}
        {}
        generator::Generator<Packet const> packets() override ;
        std::string_view name() const noexcept override ;
       private:   
      };
       class MessageHeader : private Chunk {
        Source::Ptr source_;
       public:   MessageHeader() = default;
        static constexpr Length HEADER_SIZE{};
        Length length() const noexcept ;
        export_t export_time() const noexcept ;
        sequence_t sequence() const noexcept ;
        obdomain_t observation_domain() const noexcept ;
        void  source() const noexcept ;
       private:   struct special_ {};
        static constexpr special_ special = {};
        MessageHeader( Chunk const & c, Source::Ptr source) noexcept     : Chunk{
    std::move(c)}
     , source_{
    std::move(source)}
        {}
        friend class Message;
      };
       constexpr version_t IPFIX_VERSION = 0x000a;
       class Message : public MessageHeader {};
       inline std::ranges::view auto MessageView() ;
       }
        template <typename CharT> struct ipfix::fmt::formatter<ipfix::Message, CharT> :   public ipfix::base_formatter<CharT> {
         auto format(ipfix::Message const & t, auto & ctx) const {
          return ipfix::fmt::format_to(       ctx.out(), "",       t.observation_domain(), t.length(),       t.export_time(), t.sequence());
        }
       };
        namespace ipfix {}
        namespace ipfix {
       class FieldSpecifier {
        InfoElementID id_;
        Length length_;
       public:   
        explicit constexpr FieldSpecifier(     InfoElementID id, Length length) noexcept     : id_{
    id}
     , length_{
    length}
        {}
         void  chunk_read()   ;
         constexpr InfoElementID id() const noexcept {
         return id_;
       }
         constexpr Length size() const noexcept {
         return length_;
       }
         void  is_varlen() const noexcept ;
      };
       }
        namespace ipfix {
       class FieldInfo {
        Length length_;
        InfoElement::Ptr ie_;
       public:   FieldInfo(FieldSpecifier const & spec, InfoModel const & model)     : length_{}
     , ie_{
    model.find(spec.id())}
        {}
         void  ie() const noexcept ;
         Length size() const noexcept ;
         void  is_varlen() const noexcept ;
      };
       }
        namespace ipfix {
       class TemplateSpecification {};
       class Template {
        struct TemplateBase;
        std::shared_ptr<TemplateBase> data_;
        using context_id_t = std::uint64_t;
        static inline constinit std::atomic<context_id_t> context_count;
       public:   constexpr Template() = default;
        
        Template(     TemplateID id,     TemplateSpecification && fields,     InfoModel model 
     )     : data_{
    std::make_shared<TemplateBase>(       std::move(model), id, std::move(fields))}
        {}
         explicit operator bool() const noexcept ;
         void  size() const noexcept ;
         void  is_withdrawal() const noexcept ;
         generator::Generator<     std::pair<Chunk, std::reference_wrapper<FieldInfo>>>   iter_fields() const   ;
         void    iter_record_chunks() const   ;
         std::vector<FieldInfo> const & fields() const ;
       private:   struct TemplateBase {
         TemplateID id_;
         InfoModel model_;
         TemplateSpecification spec_;
         length_t minimum_length_;
         std::vector<length_t> field_index_;
         std::vector<FieldInfo> field_infos_;
         TemplateBase(       InfoModel && model,       TemplateID id,       TemplateSpecification && spec)       : id_{
   id}
    , model_{
   std::move(model)}
    , spec_{
   std::move(spec)}
         {}
         void build_index() ;
          length_t record_length() const ;
          void      iter_record_chunks() const     ;
       };
      };
       }
        template <> struct ipfix::fmt::formatter<ipfix::Template, char>   : public ipfix::base_formatter<char> {
         auto format(ipfix::Template const & t, auto & ctx) const {
          auto out = ctx.out();
          std::size_t idx = 0;
          auto iter = t.fields();
          for (auto i = iter.begin();
      i != iter.end();
      ++i, ++idx) ;
          return out;
        }
       };
        namespace ipfix {
       constexpr setid_t TEMPLATE_SET = 2;
       class Set : public Chunk {
        MessageHeader message_header_;
        void  is_template_set(setid_t id) noexcept ;
       public:   Set(MessageHeader const & message_header, Chunk const & data) :     Chunk{
    data}
     ,     message_header_{
    message_header}
        {}
        setid_t set_id() const noexcept ;
        void  is_template_set() const noexcept ;
        void  is_options_template_set() const noexcept ;
        void  is_templatelike_set() const noexcept ;
        void  is_data_set() const noexcept ;
        Length length() const noexcept ;
        void  contents() const noexcept ;
        void  message_header() const noexcept ;
      };
       }
        template <> struct ipfix::fmt::formatter<ipfix::Set, char> :   public ipfix::base_formatter<char> {
         auto format(ipfix::Set const & t, auto & ctx) const {
          char const * type = "Unknown";
          return ipfix::fmt::format_to(       ctx.out(), "", type, t.length());
        }
       };
        namespace ipfix {
       template <typename Iter> struct PresentationOutput {
        template <typename T>   Iter operator()(Iter out, T const & v) const ;
      };
       class List : protected Chunk {
       protected:   List()    ;
      };
       template <std::derived_from<List> T> struct convert::from_chunk<T> {};
       }
        namespace ipfix {
       class BasicList : public List {};
       class Record;
       class RecordList : public List {
       public:   using List::List;
        virtual std::uint16_t template_count() const = 0;
        virtual std::uint16_t record_count() const = 0;
        virtual generator::Generator<Record> iter_records() const = 0;
        virtual generator::Generator<std::pair<Template,                                          generator::Generator<Record>>>   iter() const = 0;
      };
       class SubTemplateList : public RecordList {};
       class SubTemplateMultiList : public RecordList {};
       class Field {
        Chunk chunk_;
        MessageHeader header_;
        FieldInfo info_;
       public:   DataType datatype() const ;
        template <typename Iter>   Iter format_to(Iter out) const {
         return fmt_to<Iter>{}
    (datatype().value(), out, *this);
       }
       private:   template <typename... Args>   Field() :     chunk_{}
     ,     header_{}
     ,     info_{}
        {}
        template <typename Iter>   struct fmt_to_fn {
         template <std::size_t I>     struct fn {
          Iter operator()(Iter & out, Field const & f) const {
           auto conv = convert::from_chunk<defaults::element_t<I>>{};
           constexpr bool simple = requires (decltype(conv) fn, Chunk c) {
             fn(c);
   };
           if constexpr (simple) 
             return PresentationOutput<Iter>{}
  (out, conv(f.chunk_));
           
   else 
             return PresentationOutput<Iter>{}
  (out, conv);
           
         }
        };
       };
        template <typename Iter>   using fmt_to =     utils::IndexedCaller<defaults::types::size, fmt_to_fn<Iter>::template fn,                          Iter, Iter &, Field const &>;
        friend class Record;
      };
       }
        namespace ipfix {
       class Record {
        Chunk chunk_;
        Template template_;
        MessageHeader header_;
       public:   
        Record(Chunk chunk, Template t, MessageHeader header)     : chunk_{
    std::move(chunk)}
     ,       template_{
    t}
     ,       header_{
    header}
        {}
           generator::Generator<Field> iter() const ;
      };
       }
        template <> struct ipfix::fmt::formatter<ipfix::Record, char>   : ipfix::opt_formatter {
         auto format(ipfix::Record const & v, auto & ctx) const {
          auto out = ctx.out();
          auto fields = v.iter();
          bool first = true;
          for (auto && field : fields) 
           if (first)  
           out = field.format_to(out);
         
          return out;
        }
       };
        namespace ipfix {
       class TemplateMap {};
       class TemplateManager {
        mutable std::unordered_map<obdomain_t, TemplateMap> map_;
        InfoModel model_;
       public:   TemplateManager(InfoModel model) : model_{
    std::move(model)}
      {}
        void  infomodel() const noexcept ;
        void  templates() const ;
      };
       }
        namespace ipfix {
       class ReadEngine {
        Source::Ptr source_;
        generator::Generator<Record> next_record_;
        utils::CallbackList<Message const &> message_callbacks_;
        utils::CallbackList<Set const &> set_callbacks_;
        utils::CallbackList<Template const &> template_callbacks_;
       public:   ReadEngine(Source::Ptr source)     : source_{
    std::move(source)}
     ,       next_record_{
    record_generator()}
        {}
        template <typename T>   auto add_message_callback(T && cb) {
         return message_callbacks_.add(std::move(cb));
       }
        template <typename T>   auto add_set_callback(T && cb) {
         return set_callbacks_.add(std::move(cb));
       }
        template <typename T>   auto add_template_callback(T && cb) {
         return template_callbacks_.add(std::move(cb));
       }
        auto begin() noexcept {
         return next_record_.begin();
       }
        auto end() const noexcept {
         return next_record_.end();
       }
        void    set_generator()   ;
        generator::Generator<Record>   record_generator()   ;
      };
       }
        using std::filesystem::path;
        using namespace ipfix;
        int main(int argc, char **argv) {
         InfoModel model;
         auto reverse = [](InfoElement const & ie) {
          return reverse_standard_ie(ie);
        };
         for (int i = 2;
       i < argc;
       ++i) 
          model.read_xml(path{}
     , reverse);
        
         auto source =     std::make_shared<FileSource>(std::filesystem::path{
     argv[1]}
      , model);
         auto log_message = [](const Message &message) {
          std::cout << fmt::format("", message) << std::endl;
        };
         auto log_set = [](const Set & set) {
          std::cout << fmt::format("  ", set) << std::endl;
        };
         auto log_template = [](const Template & tmpl) {
          std::cout << fmt::format("    ", tmpl) << std::endl;
        };
         ReadEngine engine{
     source};
         engine.add_message_callback(log_message);
         engine.add_set_callback(log_set);
         engine.add_template_callback(log_template);
         for (auto & record : engine) 
          std::cout << fmt::format("", record) << std::endl;
        
       }
